#!/usr/bin/python
from rank import *
from numpy import *
from manifold import *
import cmath
from krawczyk import *
import snappy
import sys
import interval
import complex

pi = interval.interval.pi()

def iarg(z):
	temp = atan2(z.imag.inf, z.imag.sup, z.real.inf, z.real.sup)
	return interval.interval(temp[0],temp[1])


def get_data(M):
	app_sol = [] #approximated solution, [z1.real, z1.imag, z2.real, z2.imag, ....]
	ind_mat = [] #store matrix of gluing equations
	# store equations of form 'rect'
	rect = []
	equations_rect = M.gluing_equations(form = 'rect')
	chosen_equation = [] # the i_th line we chose is chosen_equation[i]     
        num_tet = M.num_tetrahedra()
        num_cusps = M.num_cusps()
        # if a cusp is complete, we do not need consistency equation for longitude.
	ignore = 0
	for n in range(num_cusps):
                rect.append(equations_rect[num_tet + n + ignore])
		chosen_equation.append(num_tet + n + ignore)
		if M.cusp_info('complete?')[n]:
                        ignore += 1
        for n in range(num_tet):
                rect.append(equations_rect[n])
                chosen_equation.append(n)
	
	for sol in M.tetrahedra_shapes('rect'):
		app_sol.append(sol.real)
		app_sol.append(sol.imag)
	
	for i in chosen_equation:
		line = equations_rect[i]
		temp = []
		for a in range(len(line)):
			if a%3<2:
				for b in line[a]:
					temp.append(float(b))
			else:
				temp.append(float(line[a]))
		ind_mat.append(temp[:])
	return app_sol[:], ind_mat[:], chosen_equation[:]

def verify_hyperbolicity(M, print_data = False):
	# first check if M has positive solution.
	if min_imaginary_shapes(M) < 0.0000000001:
		if print_data:
			print "Approximated solution is not positive enough."
		return False, []
	# Start main part
	# app_sol is the approximated solution by SnapPea
	# ind_mat is the index matrix
	# We chose n equations, i-th equation of our n equations is chosen_equation_all[i].
	app_sol, ind_mat, chosen_equation_all  = get_data(M)
	row_num = M.num_cusps()
	length_c = M.num_tetrahedra()
	if print_data:
		print "appximated solution", app_sol
		print "index matrix", ind_mat
	x0 = array(app_sol)
	ind_mat = array(ind_mat)
	row_ind = arange(row_num)
	rowend = shape(ind_mat)[0]


	# Index selection by calculating matrix rank
	for i in range(row_num,rowend+1):
		if myrank(ind_mat[:i+1,:-1],1e-8) > row_num:
			row_ind = append(row_ind,i)
			row_num += 1
		if row_num == length_c:
			break
	
	chosen_equation = []
	for i in range(len(row_ind)):
		chosen_equation.append(chosen_equation_all[i])
	ind_mat = ind_mat[row_ind,:]
	# Index matrix a, b and vector c
	a = ind_mat[:,:length_c]
	b = ind_mat[:,length_c:2*length_c]
	c = ind_mat[:,-1]
	if print_data:
		print 'approximate solution:'
		print x0
	list_abc = append(a.flatten(),b.flatten())
	list_abc = append(list_abc,c)
	list_abc = append(list_abc,x0)
	list_abc = list_abc.tolist()
	x = array(krawczyk(length_c,list_abc))
	if print_data:
		if x[-1] == 1:
			print 'verified solution:'
			print reshape(x[:-1],(length_c,4))
		else:
			print 'verification is failed'
	ans = [None]*length_c
	if x[-1] != 1:
		return False, ans
	for a in range(length_c):
		tmp_real = interval.interval(x[4*a],x[4*a+1])
		tmp_imag = interval.interval(x[4*a+2],x[4*a+3])
		ans[a] = complex.complex(tmp_real,tmp_imag)
	# check if the solutions are all positive
	for z in ans:
		if z.imag.inf<0 or (1/(1-z)).imag.inf<0 or ((z-1)/z).imag.inf<0:
			return False, ans
	if print_data:
		print "all positively oriented tetrahedra, rigorously ensured"

	# compute the sum of arguments of the solution
	result = 0

	#"""
	# check the argument of each gluing equation.
	#
	log_eq = M.gluing_equations('log')
	n = len(ans)
	ans_arg = [ans[:],ans[:],ans[:]]
	arg_verified = True
	#"""
	for j in range(n):
			#z = ans[j].real.sup + ans[j].imag.sup*1j
			z = ans[j]
			ans_arg[0][j] = iarg(z)
			ans_arg[1][j] = iarg(1/(1-z))
			ans_arg[2][j] = iarg((z-1)/z)
	for i in range(len(chosen_equation)):
		result = 0 # initialize
		for j in range(n):
			result = result + log_eq[chosen_equation[i],3*j] * ans_arg[0][j] + log_eq[chosen_equation[i],3*j+1] * ans_arg[1][j] + log_eq[chosen_equation[i],3*j +2] * ans_arg[2][j]
		result = result/pi 
		#"""
	#"""
		# result should be 
		# 1) 0 if the equation corresponds to a cusp equation (equation for the completeness of a cusp)
		# 2) 2 otherwise (equation for a Dehn filling or an edge)
		if i < M.num_cusps() and M.cusp_info('complete?')[i]: #see print_Moser_data for our way to choose equations.
			if result.inf > -0.01 and result.sup < 0.01: # if 0 is in result
				if print_data:
					print "argument condition checked, equation",i, log_eq[chosen_equation[i],:], result
			else:
				if print_data:
					print "False, equation", i, log_eq[chosen_equation[i],:], result
				arg_verified = False
		else:
			if result.inf > 1.99 and result.sup < 2.01: # if 2.0 is in result
				if print_data:
					print "argument condition checked, equation",i, log_eq[chosen_equation[i],:], result
			else:
				if print_data:
					print "False, equation", i, log_eq[chosen_equation[i],:], result
				arg_verified = False
	#"""

	return (x[-1] == 1) and arg_verified, ans[:]

